
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* Experimental code for notifying external programs of status changes 
   and the like. Will probably be moved over to dbus as soon as I (or 
   someone else) figures out the low-level interface */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_active_state.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_error.h"
#include "../../src/initng_plugin.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_control_command.h"

#include "../../initng-paths.h"

#include "initng_ngcs.h"

static void accepted_client(f_module_h * from);
static void closesock(void);

/* static int sendping(void); */
static int open_socket(void);
static void check_socket(void);
static void data_ready(f_module_h * from);
static void closeconn(ngcs_conn * conn);
static void closeconn_safe(ngcs_conn * conn);
void register_ngcs_cmds(void);
void unregister_ngcs_cmds(void);
void clean_dead_conns(void);

/*
   In the Linux implementation, sockets which are visible in the filesystem
   honour the permissions of the directory they are in.  Their owner, group
   and their permissions can be changed.  Creation of a new socket will
   fail if the process does not have write and search (execute) permission
   on the directory the socket is created in.  Connecting to the socket
   object requires read/write permission.  This behavior differs from  many
   BSD derived systems which ignore permissions for Unix sockets.  Portable
   programs should not rely on this feature for security.
 */


/* globals */
struct stat sock_stat;
const char *socket_filename;
ngcs_conn ngcs_conns;
ngcs_conn ngcs_dead_conns;
ngcs_cmd ngcs_cmds;

f_module_h fdh = { &accepted_client, -1 };

static void closesock(void)
{
    /* Check if we need to remove hooks */
    if (fdh.fds < 0)
        return;
    D_("closesock %d\n", fdh.fds);

    /* close socket and set to 0 */
    close(fdh.fds);
    fdh.fds = -1;
}

/* called by fd hook, when data is no socket */
void accepted_client(f_module_h * from)
{
    int newsock;
    struct timeval tv;
    ngcs_conn *conn;

    /* In case of lots of short-lived connections */
    clean_dead_conns();

    /* make a dumb check */
    if (from != &fdh)
        return;

    D_("Got here from fd hook.\n");
    /* we try to fix socket after every service start
       if it fails here chances are a user screwed it
       up, and we shouldn't manually try to fix anything. */
    if (fdh.fds <= 0)
    {
        /* socket is invalid but we were called, call closesocket to make sure it wont happen again */
        F_("accepted client called with fdh.fds %d, report bug\n", fdh.fds);
        closesock();
        F_("Attempting to reopen socket\n");
        open_socket();
        return;
    }

    /* create a new socket, for reading */
    if ((newsock = accept(fdh.fds, NULL, NULL)) > 0)
    {
        D_("read socket open, now setting options\n");
        tv.tv_sec = 2;
        tv.tv_usec = 0;
        if (setsockopt(newsock, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)))
            W_("Setting SO_RCVTIMEO failed!\n");
        if (setsockopt(newsock, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv)))
            W_("Setting SO_SNDTIMEO failed!\n");
        conn = (ngcs_conn *) malloc(sizeof(ngcs_conn));
        if (conn == NULL)
        {
            F_("Couldn't allocate ngcs_conn!");
            close(newsock);
            return;
        }
        conn->fdw.fds = newsock;
        conn->fdw.call_module = data_ready;
        conn->nextid = 1;
        conn->list.next = 0;
        conn->list.prev = 0;
        list_add(&conn->list, &ngcs_conns.list);
        INIT_LIST_HEAD(&conn->chans.list);
        initng_plugin_hook_add(&g.FDWATCHERS, 30, &conn->fdw);
        return;
    }

    /* temporary unavaible */
    if ((errno == EAGAIN) || (errno == EWOULDBLOCK))
    {
        W_("errno = EAGAIN!\n");
        return;
    }

    /* This'll generally happen on shutdown, don't cry about it. */
    D_("Error accepting socket %d, %s\n", fdh.fds, strerror(errno));
    closesock();
    return;
}

void data_ready(f_module_h * from)
{
    int sock = from->fds;
    int chanid, cnt;
    ngcs_cmd *cmd;
    ngcs_conn *conn = (ngcs_conn *) from;
    ngcs_data *data;
    ngcs_data sdata[3];
    ngcs_channel *chan;

    if (ngcs_readmsg(sock, &chanid, &data, &cnt))
    {
        closeconn(conn);
        return;
    }
    if (chanid == 0)
    {
        if (cnt < 0)
        {
            closeconn(conn);
            return;
        }
        else if (cnt == 0)
        {
            if (ngcs_writemsg(sock, 0, NULL, 0))
            {
                closeconn(conn);
                return;
            }
        }
        else if (data[0].type != NGCS_TYPE_STRING)
        {
            ngcs_freedata(data, cnt);
            sdata[0].type = NGCS_TYPE_INT;
            sdata[0].d.i = 0;
            sdata[1].type = NGCS_TYPE_STRING;
            sdata[1].size = -1;
            sdata[1].d.s = (char *) "Bad request";
            if (ngcs_writemsg(sock, 0, sdata, 2))
                closeconn(conn);
            return;
        }
        else
        {
            chan = malloc(sizeof(ngcs_channel));
            if (!chan)
            {
                F_("Couldn't allocate ngcs_chan!\n");
                sdata[0].type = NGCS_TYPE_INT;
                sdata[0].d.i = 0;
                sdata[1].type = NGCS_TYPE_STRING;
                sdata[1].size = -1;
                sdata[1].d.s = (char *) "Internal error";
                if (ngcs_writemsg(sock, 0, sdata, 2))
                    closeconn(conn);
                ngcs_freedata(data, cnt);
                return;
            }
            chan->id = conn->nextid++;
            while_ngcs_cmds(cmd)
            {
                if (strcmp(cmd->name, data[0].d.s) == 0)
                {
                    sdata[0].type = NGCS_TYPE_INT;
                    sdata[0].d.i = chan->id;
                    if (ngcs_writemsg(sock, 0, sdata, 1))
                    {
                        closeconn(conn);
                        free(chan);
                        ngcs_freedata(data, cnt);
                        return;
                    }
                    chan->cmd = cmd;
                    chan->conn = conn;
                    chan->user_data = NULL;
                    chan->list.next = 0;
                    chan->list.prev = 0;
                    list_add(&chan->list, &conn->chans.list);
                    if (cmd->connect)
                        cmd->connect(chan, data, cnt);
                    ngcs_freedata(data, cnt);
                    return;
                }
            }
            F_("Unknown ngcs command: %s", data[0].d.s);
            sdata[0].type = NGCS_TYPE_INT;
            sdata[0].d.i = 0;
            if (ngcs_writemsg(sock, 0, sdata, 1))
                closeconn(conn);
            free(chan);
            ngcs_freedata(data, cnt);
            return;
        }
    }
    else
    {
        while_ngcs_chans(chan, conn) if (chan->id == chanid)
        {
            if (chan->cmd->gotdata)
                chan->cmd->gotdata(chan, data, cnt);
            if (conn->fdw.fds < 0)
                closeconn(conn);
            ngcs_freedata(data, cnt);
            return;
        }
        W_("Got data on unknown channel %i", chanid);
        if (ngcs_writemsg(conn->fdw.fds, chanid, NULL, -1))
            closeconn(conn);
        ngcs_freedata(data, cnt);
        return;
    }
}

void closeconn(ngcs_conn * conn)
{
    ngcs_channel *chan, *nextchan;

    if (conn->fdw.fds >= 0)
    {
        initng_plugin_hook_del(&g.FDWATCHERS, &(conn->fdw));
        close(conn->fdw.fds);
    }
    list_for_each_entry_prev_safe(chan, nextchan, &conn->chans.list, list)
    {
        if (chan->cmd->free)
            chan->cmd->free(chan);
        list_del(&chan->list);
        free(chan);
    }
    list_del(&conn->list);
    free(conn);
}

void closeconn_safe(ngcs_conn * conn)
{
    if (conn->fdw.fds >= 0)
    {
        initng_plugin_hook_del(&g.FDWATCHERS, &(conn->fdw));
        close(conn->fdw.fds);
        conn->fdw.fds = -1;
    }
    list_move(&conn->list, &ngcs_dead_conns.list);
}

void clean_dead_conns(void)
{
    ngcs_conn *curr, *tmp;

    while_ngcs_dead_conns_safe(curr, tmp)
    {
        closeconn(curr);
    }
}

void ngcs_close_channel(ngcs_channel * chan)
{
    if (chan->conn->fdw.fds >= 0)
        if (ngcs_writemsg(chan->conn->fdw.fds, chan->id, NULL, -1))
            closeconn_safe(chan->conn);
    if (chan->cmd->free)
        chan->cmd->free(chan);
    list_del(&chan->list);
    free(chan);
}

int ngcs_channel_send(ngcs_channel * chan, ngcs_data * data, int cnt)
{
    if (chan->conn->fdw.fds < 0)
        return 1;
    if (ngcs_writemsg(chan->conn->fdw.fds, chan->id, data, cnt))
    {
        closeconn_safe(chan->conn);
        return 1;
    }
    return 0;
}

void ngcs_reg_cmd(ngcs_cmd * cmd)
{
    cmd->list.prev = 0;
    cmd->list.next = 0;
    list_add(&cmd->list, &ngcs_cmds.list);
}

#if 0
/* Send a ping to ourselfes to check if we're 100% ok. */
static int sendping()
{
    int client;
    struct sockaddr_un sockname;
    data_header head;

    D_("Sending ping\n");


    /* Create the socket. */
    client = socket(PF_UNIX, SOCK_STREAM, 0);
    if (client < 0)
    {
        F_("Failed to init socket\n");
        return FALSE;
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;

    strcpy(sockname.sun_path, socket_filename);

    /* Try to connect */
    if (connect
        (client, (struct sockaddr *) &sockname,
         (strlen(sockname.sun_path) + sizeof(sockname.sun_family))) < 0)
    {
        close(client);
        return FALSE;
    }

    D_("Sending PING..\n");
    if (ngcs_sendmsg(NGCS_PING, 0, 123, NULL, 0)) ;
    {
        F_("Unable to send PING!\n");
        close(client);
        return (FALSE);
    }
    D_("PING sent..\n");

    /* Accept "server side" */
    accepted_client(&fdh);
    /* FIXME: won't work anymore */

    D_("Reading PONG..\n");
    if (ngcs_recvall(client, &head, sizeof(head)))
    {
        F_("Unable to receive PONG!\n");
        close(client);
        return (FALSE);
    }

    D_("Got pong\n");

    return TRUE;
}
#endif

/* this will try to open a new socket */
static int open_socket()
{
    /*    int flags; */
    struct sockaddr_un serv_sockname;

    D_("Creating " SOCKET_ROOTPATH " dir\n");

    closesock();

    /* Make /dev/initng if it doesn't exist (try either way) */
    if (mkdir(SOCKET_ROOTPATH, S_IRUSR | S_IWUSR | S_IXUSR) == -1
        && errno != EEXIST)
    {
        if (errno != EROFS)
            F_("Could not create " SOCKET_ROOTPATH
               " : %s, maby / fs not mounted read-write yet?, will retry until I succeed.\n",
               strerror(errno));
        return (FALSE);
    }

    /* chmod root path for root use only */
    if (chmod(SOCKET_ROOTPATH, S_IRUSR | S_IWUSR | S_IXUSR) == -1)
    {
        /* path doesn't exist, we don't have /dev yet. */
        if (errno == ENOENT || errno == EROFS)
            return (FALSE);

        F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
           SOCKET_ROOTPATH);
    }

    /* Create the socket. */
    fdh.fds = socket(PF_UNIX, SOCK_STREAM, 0);
    if (fdh.fds < 1)
    {
        F_("Failed to init socket (%s)\n", strerror(errno));
        fdh.fds = -1;
        return (FALSE);
    }

    /* Bind a name to the socket. */
    serv_sockname.sun_family = AF_UNIX;

    strcpy(serv_sockname.sun_path, socket_filename);

    /* Remove old socket file if any */
    unlink(serv_sockname.sun_path);

    /* Try to bind */
    if (bind
        (fdh.fds, (struct sockaddr *) &serv_sockname,
         (strlen(serv_sockname.sun_path) +
          sizeof(serv_sockname.sun_family))) < 0)
    {
        F_("Error binding to socket (errno: %d str: '%s')\n", errno,
           strerror(errno));
        closesock();
        unlink(serv_sockname.sun_path);
        return (FALSE);
    }

    /* chmod socket for root only use */
    if (chmod(serv_sockname.sun_path, S_IRUSR | S_IWUSR) == -1)
    {
        F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
           serv_sockname.sun_path);
        closesock();
        return (FALSE);
    }

    /* store sock_stat for checking if we need to recreate socket later */
    stat(serv_sockname.sun_path, &sock_stat);

    /* Listen to socket */
    if (listen(fdh.fds, 5))
    {
        F_("Error on listen (errno: %d str: '%s')\n", errno, strerror(errno));
        closesock();
        unlink(serv_sockname.sun_path);
        return (FALSE);
    }

    /* Run check : */
    /*    if (!sendping())
       {
       F_("Sendping check failed, ngcs comunication not available (if you see this open a bug)\n");
       closesock();
       return (FALSE);
       } */

    return (TRUE);
}

/* this will check socket, and reopen on failure */
static void check_socket()
{
    struct stat st;

    D_("Checking socket\n");

    /* Check if socket needs reopening */
    if (fdh.fds <= 0)
    {
        D_("fdh.fds not set, opening new socket.\n");
        open_socket();
        return;
    }

    /* stat the socket, reopen on failure */
    memset(&st, 0, sizeof(st));
    if (stat(socket_filename, &st) < 0)
    {
        W_("Stat failed! Opening new socket.\n");
        open_socket();
        return;
    }

    /* compare socket file, with the one that we know, reopen on failure */
    if (st.st_dev != sock_stat.st_dev || st.st_ino != sock_stat.st_ino
        || st.st_mtime != sock_stat.st_mtime)
    {
        F_("Invalid socket found, reopening\n");
        open_socket();
        return;
    }

    D_("Socket ok.\n");
    return;
}

/* this function, will make a check for socket, on every new service that goes up */
int service_status(active_db_h * service)
{
    /* only try open, when a service got up */
    if (IS_MARK(service, &DONE))
        check_socket();

    return (TRUE);
}

int module_init(const char *version)
{
    D_("module_init(ngcs);\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }


    /* zero globals */
    fdh.fds = -1;
    memset(&sock_stat, 0, sizeof(sock_stat));
    INIT_LIST_HEAD(&ngcs_conns.list);
    INIT_LIST_HEAD(&ngcs_dead_conns.list);
    INIT_LIST_HEAD(&ngcs_cmds.list);

    /* decide which socket to use */
    if (g.i_am_init)
        socket_filename = SOCKET_FILENAME_REAL;
    else
        socket_filename = SOCKET_FILENAME_TEST;
    D_("Socket is: %s\n", socket_filename);

    D_("adding hook, that will reopen socket, for every started service.\n");
    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 50, &service_status);
    initng_plugin_hook_add(&g.FDWATCHERS, 30, &fdh);

    register_ngcs_cmds();

    /* do the first socket directly */
    open_socket();

    D_("ngcs.so.0.0 loaded!\n");
    return (TRUE);
}

void module_unload(void)
{
    ngcs_conn *curr, *tmp;

    D_("module_unload(ngcs);\n");

    /* close open sockets */
    closesock();
    while_ngcs_conns_safe(curr, tmp)
    {
        closeconn(curr);
    }

    while_ngcs_dead_conns_safe(curr, tmp)
    {
        closeconn(curr);
    }

    unregister_ngcs_cmds();

    /* remove hooks */
    initng_plugin_hook_del(&g.FDWATCHERS, &fdh);
    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &service_status);

    D_("ngcs.so.0.0 unloaded!\n");

}
