#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_active_state.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_error.h"
#include "../../src/initng_plugin.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_control_command.h"

#include "../../initng-paths.h"
#include "initng_ngcs.h"
#include "ngcs_marshal.h"

void register_ngcs_cmds(void);
void unregister_ngcs_cmds(void);
static void ngcs_cmd_halt(ngcs_channel * chan, ngcs_data * data, int cnt);
static void ngcs_cmd_reboot(ngcs_channel * chan, ngcs_data * data, int cnt);
static int service_status_watch(active_db_h * service);
static void ngcs_cmd_watch(ngcs_channel * chan, ngcs_data * data, int cnt);
static void ngcs_free_watch(ngcs_channel * chan);
static int service_output_watch(active_db_h * service, process_h * x);

ngcs_cmd ngcs_halt_cmd = {
    "halt",
    ngcs_cmd_halt,
    NULL,
    NULL,
    {0, 0}
};

ngcs_cmd ngcs_reboot_cmd = {
    "halt",
    ngcs_cmd_reboot,
    NULL,
    NULL,
    {0, 0}
};

ngcs_cmd ngcs_watch_cmd = {
    "watch",
    ngcs_cmd_watch,
    NULL,
    ngcs_free_watch,
    {0, 0}
};

typedef struct ngcs_watch_s
{
    ngcs_channel *chan;
    char *name;
    struct list_head list;
} ngcs_watch;

ngcs_watch watches;

int service_status_watch(active_db_h * service)
{
    ngcs_watch *watch, *nextwatch;
    int n, ret = 0;
    ngcs_data *dat = NULL;

    n = ngcs_marshal_active_db_h(service, NULL, 0);
    list_for_each_entry_prev_safe(watch, nextwatch, &watches.list, list)
    {
        if (watch->name == NULL || strcmp(watch->name, service->name) == 0)
        {
            if (!dat)
            {
                dat = malloc(n * sizeof(ngcs_data));
                if (dat == NULL)
                {
                    F_("malloc() failure\n");
                    return TRUE;
                }
                ret = ngcs_marshal_active_db_h(service, dat, n);
                if (ret < 0)
                {
                    F_("ngcs_marshal_active_db_h() failed!\n");
                    free(dat);
                    return TRUE;
                }
            }
            ngcs_channel_send(watch->chan, dat, ret);
        }
    }
    if (dat)
        free(dat);
    return TRUE;
}

static int service_output_watch(active_db_h * service, process_h * x)
{
    ngcs_watch *watch, *nextwatch;
    ngcs_data dat[2];

    dat[0].type = NGCS_TYPE_STRING;
    dat[0].size = -1;
    dat[0].d.s = service->name;
    dat[1].type = NGCS_TYPE_STRING;
    dat[1].size = -1;
    dat[1].d.s = x->buffer + x->buffer_pos;
    list_for_each_entry_prev_safe(watch, nextwatch, &watches.list, list)
    {
        if (watch->name == NULL || strcmp(watch->name, service->name) == 0)
        {
            ngcs_channel_send(watch->chan, dat, 2);
        }
    }
    return FALSE;                           /* FIXME - right thing to do? */
}

void ngcs_cmd_watch(ngcs_channel * chan, ngcs_data * data, int cnt)
{
    ngcs_watch *watch = malloc(sizeof(ngcs_watch));

    if (!watch)
    {
        F_("Malloc failed!\n");
        ngcs_close_channel(chan);
        return;
    }
    watch->chan = chan;
    if (cnt < 2 || data[1].type != NGCS_TYPE_STRING)
    {
        watch->name = NULL;
    }
    else
    {
        watch->name = i_strdup(data[1].d.s);
        if (!watch->name)
        {
            free(watch);
            F_("Malloc failed!\n");
            ngcs_close_channel(chan);
            return;
        }
    }
    watch->list.prev = 0;
    watch->list.next = 0;
    list_add(&watch->list, &watches.list);
    chan->user_data = watch;
}

void ngcs_free_watch(ngcs_channel * chan)
{
    ngcs_watch *watch = chan->user_data;

    if (!watch)
        return;
    list_del(&watch->list);
    if (watch->name)
        free(watch->name);
    free(watch);
    chan->user_data = 0;
}

void register_ngcs_cmds(void)
{
    ngcs_reg_cmd(&ngcs_halt_cmd);
    ngcs_reg_cmd(&ngcs_reboot_cmd);
    ngcs_reg_cmd(&ngcs_watch_cmd);
    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 50, &service_status_watch);
    initng_plugin_hook_add(&g.PIPEWATCHERS, 50, &service_output_watch);
    INIT_LIST_HEAD(&watches.list);
}

void unregister_ngcs_cmds(void)
{
    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &service_status_watch);
    initng_plugin_hook_del(&g.PIPEWATCHERS, &service_output_watch);
    /* TODO */
}

void ngcs_cmd_halt(ngcs_channel * chan, ngcs_data * data, int cnt)
{
    (void) data;
    (void) cnt;
    g.when_out = THEN_HALT;
    initng_handler_stop_all();
    ngcs_close_channel(chan);
}

void ngcs_cmd_reboot(ngcs_channel * chan, ngcs_data * data, int cnt)
{
    (void) data;
    (void) cnt;
    g.when_out = THEN_REBOOT;
    initng_handler_stop_all();
    ngcs_close_channel(chan);
}
