
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <sys/types.h>                      /* time_t */
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>                          /* printf() */
#include <assert.h>
#include <stdarg.h>
#include <syslog.h>

#include "initng_syslog.h"
#include "../../src/initng_global.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_static_states.h"

static int syslog_fetch_output(active_db_h * service, process_h * process);
static void syslog_print_system_state(h_sys_state state);
static int syslog_print_status_change(active_db_h * service);
static void check_syslog(void);
static void initng_log(int prio, const char *owner, const char *format, ...);
static void free_buffert(void);

/* global variable */
int syslog_running;

static void check_syslog(void)
{
    struct stat st;

    if (stat("/dev/log", &st) == 0 && S_ISSOCK(st.st_mode))
    {
        syslog_running = 1;

        /* print out the buffert if any now */
        if (!list_empty(&log_list.list))
        {
            log_ent *current, *safe = NULL;

            while_log_list_safe(current, safe)
            {
                initng_log(current->prio, current->owner, "%s",
                           current->buffert);
                free(current->buffert);
                if (current->owner)
                    free(current->owner);
                list_del(&current->list);
            }
            INIT_LIST_HEAD(&log_list.list);
        }
    }
    else
    {
        syslog_running = 0;
    }
}

static void free_buffert(void)
{
    log_ent *current, *safe = NULL;

    /* give syslog a last chanche to come alive */
    check_syslog();
    while_log_list_safe(current, safe)
    {
        free(current->buffert);
        if (current->owner)
            free(current->owner);
        list_del(&current->list);
    }
    INIT_LIST_HEAD(&log_list.list);

    /* log directly to syslog from now, even if it might not exist */
    syslog_running = 1;
}

static void initng_log(int prio, const char *owner, const char *format, ...)
{
    va_list ap;

    va_start(ap, format);


    /* if syslog is running, send it directly */
    if (syslog_running == 1)
    {
        /* if owner is set, we have openlog with right owner */
        if (owner)
        {
            closelog();
            openlog(owner, 0, LOG_LOCAL1);

            vsyslog(prio, format, ap);

            closelog();
            openlog("InitNG", 0, LOG_LOCAL1);
        }
        else
            vsyslog(prio, format, ap);

    }
    else
    {
        char b[201];
        log_ent *tmp = (log_ent *) i_calloc(1, sizeof(log_ent));

        if (!tmp)
            return;

        tmp->prio = prio;
        vsnprintf(b, 200, format, ap);
        tmp->buffert = i_strdup(b);

        if (owner)
            tmp->owner = i_strdup(owner);
        else
            tmp->owner = NULL;

        list_add(&tmp->list, &log_list.list);

    }

    va_end(ap);
}

/* add values to syslog database */
static int syslog_print_status_change(active_db_h * service)
{
    assert(service);
    assert(service->name);

    if (IS_UP(service))
    {
        check_syslog();
        initng_log(LOG_NOTICE, NULL, "Service %s is started.\n",
                   service->name);
        return (TRUE);
    }

    if (IS_MARK(service, &STOPPED))
    {
        initng_log(LOG_NOTICE, NULL, "Service %s has been stopped.\n",
                   service->name);
        return (TRUE);
    }
    if (IS_MARK(service, &FAIL_STARTING))
    {
        initng_log(LOG_NOTICE, NULL, "Service %s FAILED to start.\n",
                   service->name);
        return (TRUE);
    }

    if (IS_MARK(service, &FAIL_STOPPING))
    {
        initng_log(LOG_NOTICE, NULL, "Service %s FAILED to stop.\n",
                   service->name);
        return (TRUE);
    }

    /* leave */
    return (TRUE);
}

static void syslog_print_system_state(h_sys_state state)
{
    switch (state)
    {
        case STATE_UP:
            initng_log(LOG_NOTICE, NULL, "System is up and running!\n");
            /*
             * if syslogd have not been started yet, its no ida,
             * to spend memory for the buffert anyway.
             */
            free_buffert();
            return;
        case STATE_STARTING:
            initng_log(LOG_NOTICE, NULL, "System is starting up.\n");
            return;
        case STATE_STOPPING:
            initng_log(LOG_NOTICE, NULL, "System is going down.\n");
            return;
        case STATE_ASE:
            initng_log(LOG_NOTICE, NULL, "Last service exited.\n");
            return;
        case STATE_EXIT:
            initng_log(LOG_NOTICE, NULL, "Initng is exiting.\n");
            return;
        case STATE_RESTART:
            initng_log(LOG_NOTICE, NULL, "Initng is restarting.\n");
            return;
        case STATE_HALT:
            initng_log(LOG_NOTICE, NULL, "System is halting.\n");
            return;
        case STATE_REBOOT:
            initng_log(LOG_NOTICE, NULL, "System is rebooting.\n");
            return;
        default:
            return;
    }
    return;
}

static int syslog_fetch_output(active_db_h * service, process_h * process)
{
    char log[201];
    int pos = process->buffer_pos;
    int i;

    assert(service);
    assert(service->name);

    /* print every line, ending with a '\n' as an own syslog */
    while (process->buffer[pos])
    {
        i = 0;
        /* count the number of char before '\n' */
        while (process->buffer[pos + i] && process->buffer[pos + i] != '\n'
               && i < 200)
            i++;

        /* copy that many chars to our temporary log array */
        strncpy(log, &process->buffer[pos], i);
        log[i] = '\0';

        /* send it to syslog */
        initng_log(LOG_NOTICE, service->name, "%s", log);

        /* step forward */
        pos += i;

        /* and skip the newline if any */
        if (process->buffer[pos])
            pos++;

    }
    return (TRUE);
}


static int syslog_print_error(e_mt mt, const char *file, const char *func,
                              int line, const char *format, va_list arg)
{
    assert(file);
    assert(func);
    assert(format);
    char tempspace[200];

    vsnprintf(tempspace, 200, format, arg);

    switch (mt)
    {
        case MSG_FAIL:
#ifdef DEBUG
            syslog(LOG_EMERG, "\"%s\", %s() #%i FAIL: %s", file, func, line,
                   tempspace);
#else
            syslog(LOG_EMERG, "FAIL: %s", tempspace);
#endif
            return (TRUE);
        case MSG_WARN:
#ifdef DEBUG
            syslog(LOG_WARNING, "\"%s\", %s() #%i WARN: %s", file, func, line,
                   tempspace);
#else
            syslog(LOG_EMERG, "WARN: %s", tempspace);
#endif
            return (TRUE);
        default:
            syslog(LOG_NOTICE, "%s", tempspace);
            return (TRUE);
    }

    /* newer get here */
    return (FALSE);
}




int module_init(const char *version)
{
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    /* Dont clutter syslog in fake mode */
    if (getpid() != 1 || !g.i_am_init)
    {
        W_("Pid is not 1, (%i), or g.i_am_init not set and the syslog plugin won't load when running in fake mode, to prevent cluttering up the log-files.\n", getpid());
        return (TRUE);
    }
    D_("Initziating syslog plugin\n");

    INIT_LIST_HEAD(&log_list.list);
    check_syslog();

    setlogmask(LOG_UPTO(LOG_NOTICE));
    openlog("InitNG", 0, LOG_LOCAL1);

    initng_plugin_hook_add(&g.ERR_MSG, 50, &syslog_print_error);
    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 100, &syslog_print_status_change);
    initng_plugin_hook_add(&g.SWATCHERS, 100, &syslog_print_system_state);
    initng_plugin_hook_add(&g.PIPEWATCHERS, 100, &syslog_fetch_output);

    return (TRUE);
}


void module_unload(void)
{
    /* Dont clutter syslog in fake mode */
    if (!g.i_am_init)
    {
        D_("The syslog plugin won't load when running in fake mode, to prevent cluttering up the log-files.\n");
        return;
    }

    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &syslog_print_status_change);
    initng_plugin_hook_del(&g.SWATCHERS, &syslog_print_system_state);
    initng_plugin_hook_del(&g.PIPEWATCHERS, &syslog_fetch_output);
    initng_plugin_hook_del(&g.ERR_MSG, &syslog_print_error);
    free_buffert();
    closelog();
}
